"""This single file is needed to build the Client development dashboard."""

from promptmodel import Client

# Example imports
# from <dirname> import <objectname>

app = Client()

# Example usage
# This is needed to integrate your codebase with the prompt engineering dashboard
# app.include(<objectname>)
